#!/usr/bin/perl

use SOAP::Lite; 

#use SOAP::Lite +trace => [qw (debug)];

#Defining Signal handler for HUP signal.
$SIG{HUP} = sub {
                        &logMessage(1,"Recieved HUP Signal");
                        if ($CHILD_PID != $$)
                        {
                                 kill 1, $CHILD_PID;
                        }
                        do $CONFIGURATION_FILE;
		};

$CONF_WS_LOG_LEVEL = 0;
$gv_service; 
sub main
{
	&beDaemon if($CONF_DEBUG == 0);
	while (1)
	{
		&setConfParameter;
		if ( &checkAnotherInstance > 0 )
		{
			my $PID = fork();
			if ($PID > 0)
			{
				#this is parent process and it is monitor process
				$CHILD_PID = $PID;
				$SIG{TERM} = sub {kill 15,$CHILD_PID; die;};
				waitpid($CHILD_PID, WNOHANG);
				sleep($CONF_RETRY_WAIT * 60); 
			}
			else
			{
				$CHILD_PID = $$;
				$SIG{TERM} = 'DEFAULT'; 
				&loopForever;
			}	
		}
		else {last;}
	}			 
}  

sub setConfParameter
{
	##############################################
	#                                            #
	#    Setting the Configuration From File     #
	#                                            #
	##############################################         

	$CONF_BASE_LOG_FILE_NAME	= $Base_Logfile if(defined $Base_Logfile);
	$CONF_MAX_SLEEP_SECONDS 	= $Max_Sleep_Seconds if(defined $Max_Sleep_Seconds);
	$CONF_UNZIP_TMP_DIR 		= $Tmp_Dir if(defined $Tmp_Dir);
	$CONF_STATEFILE 		= $Publisher_Statefile if(defined $Publisher_Statefile);
	$CONF_SERVER_WSDL 		= $Gridview_Server_WSDL if(defined $Gridview_Server_WSDL); 
	$CONF_LOCKFILE 			= $Publisher_Lockfile if(defined $Publisher_Lockfile);
	$CONF_LOGFILE 			= $Publisher_Logfile if(defined $Publisher_Logfile);
	$CONF_TIMEOUT 			= $Publisher_Dead_Timeout if(defined $Publisher_Dead_Timeout);
	$CONF_RETRY_WAIT 		= $Publisher_Retry_Wait if(defined $Publisher_Retry_Wait);
	$CONF_LOG_LEVEL 		= $Publisher_Log_Level if(defined $Publisher_Log_Level);
	&logMessage(1,"setConfParameter: done config");
}


#To check other Instance is running or hung.
#If it is not running or hung then returns 1 else 0
sub checkAnotherInstance  
{
	my %lock = &readLock ($CONF_LOCKFILE);
	if (length $lock{PID} <= 0)
	{
		sleep 3;
		%lock = &readLock ($CONF_LOCKFILE);
	}
	if(length $lock{PID} >0 )
	{
		my $cmd = "ps -p $lock{PID} | grep " .substr($0,rindex($0,'/')+1,14);
		my $out = readpipe ("$cmd");
		return 1 if (length ($out) <= 0 );
		if ($lock{LAST_TIME} > time - $CONF_TIMEOUT * 60){
			&logMessage (1, "checkAnotherInstance: Another Instance is already running with PID=$lock{PID} Exiting... ");
			return 0;
		}
		else 
		{
			&logMessage (1, "checkAnotherInstance: Killing the Hung process PID=$lock{PID} ...");
			kill (9, $lock{PID}) if (length $out > 0);
			return 1;
		}
	}
	return 1;
}

#This function make this program a daemon process. 
sub beDaemon
{
        use POSIX qw(setsid);

        chdir '/'                 or die;
        umask 0;
        open STDIN, '/dev/null'   or die ;
        open STDOUT, '/dev/null' or die ;
        open STDERR, '/dev/null' or die ;
        defined(my $pid = fork)   or die;
        exit if $pid;
        setsid                    or die;
}

#This is to make the Publisher remain in infinite loop
#sleep after publishing the records for sometime and 
#recheck the availability of new records.
sub loopForever
{
	&logMessage (1, "loopForever: Starting with PID=$$ ...");
	$gv_service = SOAP::Lite->service($CONF_SERVER_WSDL);
	my $ws_log_level_read_time;
	while (1)
	{
		&setConfParameter;
		if($ws_log_level_read_time < time() - 1800) {
			&readWSLogLevel;
			$ws_log_level_read_time = time;
		}
		my $starttime = time;
		&writeLock($CONF_LOCKFILE);
		&parseAndPublish($CONF_BASE_LOG_FILE_NAME);
		my $endtime = time;
		my $sleep_time = $CONF_MAX_SLEEP_SECONDS - ($endtime - $starttime);
		$sleep_time = 0 if ($sleep_time <0);
		&logMessage (1, "loopForever: Going to sleep for $sleep_time");
		sleep $sleep_time;  
	}
}

#This is to handle the Log rotated file. It unzips and deletes the 
#Log Files rotated and zipped by LogRotate.
sub handleLogRotateFile
{
	my $file_name = shift;
	my $index = shift;
	my $operation = shift;
	
	my $local_name = $file_name . "." . $index . ".gz";
	my $r_index = rindex($file_name,"/");
	my $dst_file = substr($file_name, rindex($file_name,"/")+1);
	my $output = $CONF_UNZIP_TMP_DIR . "/" . $dst_file . "." . $index;

	if ($operation == 0)
	{
		&logMessage (2, "handleLogRotateFile: Unzipping File $local_name to $output");
		my $str = "gunzip -c $local_name > $output 2>/dev/null";
		system ($str);
		return $output;
	}
	if ($operation == 1)
	{
		&logMessage (2, "handleLogRotateFile: Deleting File $output");
		my $str = "rm -f $output";
		system ($str);
		return $output;
	}
	
}	

#This reads the configuration file and detects from where onwards new 
#records have to be published. It checks the rotated files also.
sub parseAndPublish
{
	my $file = shift;
	# Read the last State 
	my %state = &readLastState($CONF_STATEFILE);

	my $loop_exit, $backtrack = 0; #counter to check how many files to parse
	if($state{LAST_RECORD})	# If there is a Last Record means a previous atate
	{
		my $i;
		for($i =0;$i<3; $i++)
		{
			# File_name of the alernative files
			$t_file = $file;  
			if( $i != 0 ) {	$t_file = &handleLogRotateFile($file, $i, 0); }
			&logMessage (3, "parseAndPublish1: Checking File $t_file");
			open (local $TMPHANDLE, $t_file) or  &logMessage (3, "parseAndPublish1: Cannot Open File $t_file");
			#Matching the last record in the file with the state 
			#This is to check if the file is bigger than the seek size 
			$backtrack = $i;
			if(seek ($TMPHANDLE, $state{SEEK} - 1 - length($state{LAST_RECORD}), 0))
			{
				&logMessage (6, "parseAndPublish1: Seeked to Last state position in $t_file");
				$_ = <$TMPHANDLE>;
				chomp;
				my $last = $state{LAST_RECORD};
				&logMessage (5, "parseAndPublish1: Comapring State with the current LAST_RECORD=$last :::: NEW=$_");
				last if($state{LAST_RECORD} eq $_); #State matches in this File So end the Loop
				&logMessage (6, "parseAndPublish1: Comaprison Failed for File $t_file");
			}
			close ($TMPHANDLE);
			&handleLogRotateFile($file, $i, 1) if( $i != 0 ) ; 
		}
		$loop_exit = $i;
	}
	#Start form the backTracked file count to the last file 
	for ($i = $backtrack; $i>=0; $i--)
	{
		$t_file = $file; 
		if( $i != 0 ) {	$t_file = &handleLogRotateFile($file, $i, 0); }
		my $loc_seek = 0; my $write_state = 0 ;
		$loc_seek = $state{SEEK} if ($backtrack == $i); #Do Seek only for first file 
		$loc_seek = 0 if ( $loop_exit == 3 ); #Do Seek only for first file 
		$write_state = 1 if ($i == 0);
		$write_state = 1;
#		$loc_seek=0;	
		&logMessage (3, "parseAndPublish2: Going to Parse and Publish the File $t_file with seek $loc_seek");
		&publishFromFile($t_file, $loc_seek, $write_state);#Write state only for the last file
		&handleLogRotateFile($file, $i, 1) if ($i != 0); 
	}
}

sub readWSLogLevel
{
	my $soap_success = "Successfull";
        $gv_service->on_fault (sub {
                        my($soap, $res) = @_;
                        &logMessage(1,$res?$res->faultstring : $soap->transport->status );
                        $soap_success = "Unsuccessfull";
                });
	my $level = $gv_service->getWSLogLevel($CONF_TABLE_NAME, $CONF_VERSION);
	if(length($level) !=  1 && $soap_success eq "Successfull"){
		&logMessage(1,"readWSLogLevel: get unexpected response dying"); die;}
        $CONF_WS_LOG_LEVEL = $level if ($soap_success eq "Successfull");
	&logMessage(4,"First Message to WS Log Recvd Log Level= $CONF_WS_LOG_LEVEL");
	return 0;
}

sub readLock #read the Lock file for information of other instance running
{
        my $st_file = shift;
	my %ret;
        open (my $LOCKFILE,$st_file) or return 0;
        foreach (<$LOCKFILE>){
                chomp;
                $ret{PID} = $1 if (/^PID=(\d+)\s*/);
                $ret{LAST_TIME} = $1 if (/^LAST_TIME=(\d*)\s*/);
        }
        close($LOCKFILE);
	&logMessage (3, "readLock: Read Lock Values PID=$ret{PID} LAST_TIME=$ret{LAST_TIME}");
        return %ret;
}

sub readLastState #read the last state stored in the state file
{
        my $st_file = shift;
	my %ret;
        open (my $STATEFILE,$st_file) or return 0;
        foreach (<$STATEFILE>){
                chomp;
                $ret{SEEK} = $1 if (/^SEEK=(\d*)\s*/);
                $ret{LAST_RECORD} = $1 if (/^LAST_RECORD=(.*)/);
        }
        close($STATEFILE);
	&logMessage (5, "readLastState: Read Last State Values LAST_SEEK=$ret{SEEK} LAST_RECORD=$ret{LAST_RECORD}");
        return %ret;
}

sub writeLock #writes the Lock to prevent other process to start
{
        my $st_file = shift;
	my $ltime = time;
        open (my $LOCKFILE," > $st_file") or return 0;
        print $LOCKFILE "LAST_TIME=$ltime\n";
        print $LOCKFILE "PID=$$\n";
        close($LOCKFILE);
	return 1;
}

sub writeLastState #writes the LAST state to the state file
{
        my $st_file = shift;
        my $data = shift;
	my $seek = shift;
	
        open (my $STATEFILE," > $st_file") or return 0;
        print $STATEFILE "LAST_RECORD=$data\n";
        print $STATEFILE "SEEK=$seek\n";
        close($STATEFILE);
}

sub showMessage {
	my $level = shift;
	my $msg = shift;
	
	if($CONF_DEBUG > 0){
		print ("$msg\n");
		return if ($CONF_DEBUG == 1 ); 
	}
	open (local $LOGFILEHANDLE, ">> $CONF_LOGFILE" );  
	print ($LOGFILEHANDLE "$msg\n");
	close $LOGFILEHANDLE;
	return ;
} 

sub sendWSLogMessage
{	
	my $level =shift;
	my $msg = shift;	
	$gv_service->on_fault (sub {
                        my($soap, $res) = @_;
                        &logMessage(1,$res?$res->faultstring : $soap->transport->status );
                        $soap_success = "Unsuccessfull";
                });
	my $res=0;
	$res = $gv_service->putLog($CONF_TABLE_NAME, $level, length ($msg), $msg);
	if($res != length($msg) && $soap_success eq "Successfull") {
		&logMessage(1,"Failed in WS putLog dying send ". length($msg) ." received $res"); die; }
	return;
}

sub logMessage {
	my $level = shift;
	my $msgt = shift;
	my $msg = localtime() . " :$CONF_VERSION:$$: $msgt\n";
	if($level <= $CONF_LOG_LEVEL)
	{
		&showMessage($level , $msg);
	}
	if($level >= 4 and $level <= $CONF_WS_LOG_LEVEL + 3 )
	{
		# Write the message to WS also
		&sendWSLogMessage($level ,$msg);
	}
} 

#This function reads the requested log file and calls the parser 
#Checks the parser return and take action according to parser return value
#Also stores the state of last valid record published.
sub publishFromFile
{
	my $file_name = shift;
	my $seek = shift;
	my $write_state = shift;
	open (local $HANDLE, $file_name) or return &logMessage (3, "publishFromFile: Cannot Open File $file_name");
	seek($HANDLE, $seek, 0);
	my $reccount=0, @records, $no_of_lines=0, $last_pos, $cur_last_pos; 
	my $last_valid_line, $cur_last_valid_line;
	while (!eof $HANDLE) {
		$no_of_lines++;
		my $line = <$HANDLE>;
		my $tt_line = chomp $line;
		if(length($tt_line) == length($line)){
			&logMessage(5,"publishFromFile: Incomplete Line read ::$line");
			last;
		}
		&logMessage(6,"publishFromFile: Sending to Parse LINE::$line");
		my %parse_return = &parseLine($line);
		my $command = $parse_return{GRIDVIEW_PUBLISHER};
		if($command =~ /^GRIDVIEW_PUBLISHER_ERROR.*/){
			&logMessage (5, "publishFromFile: $command");
			next;
		}
		if($command =~ /^GRIDVIEW_PUBLISHER_IGNORE.*/){
			&logMessage(5, "publishFromFile: $command");
			next;
		}
		if($command =~ /^GRIDVIEW_PUBLISHER_PUBLISH.*/){
			$cur_last_pos = tell ($HANDLE);
			$cur_last_valid_line = $line;
			my $log_msg = $command;
			foreach $key (keys(%parse_return)){
				$log_msg .= "::" . $key . ":=" . $parse_return{$key}; 
				$records[$reccount]{$key}=$parse_return{$key};
			}
			&logMessage(6, "publishFromFile:$reccount: $log_msg");
			$reccount++;
			if($reccount >= 60 )
			{
				&logMessage(4, "publishFromFile: Publishing $reccount out of $no_of_lines lines");
				if(&publishTuple(\@records)){
					&logMessage(4, "publishFromFile: Successfully Published");
					$reccount = 0; $no_of_lines =0;
					$last_pos = $cur_last_pos;
					$last_valid_line = $cur_last_valid_line;
					$#records = -1;
				}
				else {last;}
			}	
			next;
		}
		&logMessage(4, "Unexpected Response from Parser $command");
	}
	if($reccount > 0 )
	{
		&logMessage(4, "publishFromFile: Publishing $reccount out of $no_of_lines lines");
		if(&publishTuple(\@records) ){
			&logMessage(4, "publishFromFile: Successfully Published");
			$reccount = 0; $no_of_lines =0;
			$last_pos = $cur_last_pos;
			$last_valid_line = $cur_last_valid_line;
			$#records = -1;
		}
	}	
	close ($HANDLE);
	if($write_state >0)
	{
		if($last_valid_line)
		{
			&logMessage (5, "publishFromFile: Writing Last State to file $CONF_STATEFILE SEEK=$last_pos LAST_RECORD=$last_valid_line"); 
			&writeLastState($CONF_STATEFILE, $last_valid_line, $last_pos);
		}
	}
}

#Function for publishing the data to Web Service Archiver
sub publishTuple
{
	my $ref = shift;
	my @tuple = @{$ref};
	my $len = @tuple;
	my $i;
	my $message = "#";
	return 0 if ($len == 0);
	for($i = 0; $i<$len; $i++)
	{
		$message .= "#";
		foreach $key (@CONF_TUPLE_ORDER)
		{
			$message .= urlEncodeData($tuple[$i]{$key}) . "|";
		}
		chop $message;
	}
	$message = &getHeader() . $message;
	#chop $message;
	#print "$message\n"; 
	&logMessage(6,"publishTuple: Publishing $len records as $message"); 
	my $soap_success = "Successfull";
	$gv_service->on_fault (sub { 
			my($soap, $res) = @_; 
			&logMessage(1,$res?$res->faultstring : $soap->transport->status );
			$soap_success = "Unsuccessfull"; 
		});
	my $res = $gv_service->putData($CONF_TABLE_NAME, $i, length ($message), $message);
	if($res != length($message) && $soap_success eq "Successfull") {
		&logMessage(1,"Failed in WS putData dying return $res length =" . length($message) ); die; }

	return 1 if ($soap_success eq "Successfull");
	return 0;
}

#To encode the delimiter in case they are part of data;
sub urlEncodeData
{
	my $string = shift;
	my $istring = $string;
	$string =~ s/%/%25/g;
	$string =~ s/#/%23/g;
	$string =~ s/\|/%7c/g;
	&logMessage (6, "urlEncodeData: Converted $istring to $string") if($istring ne $string);
	return $string;
}
